/*******************************************************************************
 *
 * hyper-channel: implements CPU side-channel based hypervisor detection.
 *
 * Uses RSB side-channel spy that puts Return Stack Buffer (RSB) structure in a known state
 * and monitors its possible corruption by a hyperviser during VMExit.
 *
 * For educational purposes only. Use it on your onw risk.
 *
 * 04/19/2008 c7zero@c7zero.info
 *            www.c7zero.info
 *
 ******************************************************************************/

#include "bsr.h"

unsigned hi1 = 0, lo1 = 0, hi2 = 0, lo2 = 0;


void __declspec(naked) func15(void)
  {
    __asm {
      pusha
      cpuid
      popa
    }

//    __asm push eax
//    __asm mov eax, cr3
//    __asm mov cr3, eax
//    __asm pop eax

    // --
    // -- start measuring 16 returns
    // --
    __RDTSC( hi1, lo1 );

    __asm ret
  }
void __declspec(naked) func14(void)
  {
    __asm call func15
    __asm ret
  }
void __declspec(naked) func13(void)
  {
    __asm call func14
    __asm ret
  }
void __declspec(naked) func12(void)
  {
    __asm call func13
    __asm ret
  }
void __declspec(naked) func11(void)
  {
    __asm call func12
    __asm ret
  }
void __declspec(naked) func10(void)
  {
    __asm call func11
    __asm ret
  }
void __declspec(naked) func9(void)
  {
    __asm call func10
    __asm ret
  }
void __declspec(naked) func8(void)
  {
    __asm call func9
    __asm ret
  }
void __declspec(naked) func7(void)
  {
    __asm call func8
    __asm ret
  }
void __declspec(naked) func6(void)
  {
    __asm call func7
    __asm ret
  }
void __declspec(naked) func5(void)
  {
    __asm call func6
    __asm ret
  }
void __declspec(naked) func4(void)
  {
    __asm call func5
    __asm ret
  }
void __declspec(naked) func3(void)
  {
    __asm call func4
    __asm ret
  }
void __declspec(naked) func2(void)
  {
    __asm call func3
    __asm ret
  }

void __declspec(naked) func1(void)
  {
    __asm call func2
    __asm ret
  }
void __declspec(naked) func0(void)
  {
    __asm call func1
    __asm ret
  }


__int64 bsr_detect( int iter_cnt )
{

  __int64 tsc_delta_avg = 0;
  __int64 cur = 0;
  int i;

  DbgPrint( "[bsr][spy] > fill RSB (16 nested CALL branches) -> CPUID\n" );

  for( i = 0; i < iter_cnt; i++ )
    {

      hi1 = lo1 = hi2 = lo2 = 0;

      __asm cli
      __asm call func0  

// empty rdtsc timing (43 clks Dothan core)
//      __RDTSC( hi1, lo1 );

      // --
      // -- stop measuring 16 returns
      // --
      __RDTSC( hi2, lo2 );

      // -- enable interrupts
      __asm sti

      // --
      // -- calculate time taken to execute 16 RET instructions
      // --
      cur = __DURATION( hi2, lo2, hi1, lo1 );
      tsc_delta_avg += cur;

    }

  tsc_delta_avg /= iter_cnt;
  DbgPrint( "[bsr][spy] < after CPUID measure 16 RET [averaged over %u iterations]: %llu clk\n", iter_cnt, tsc_delta_avg );
  return tsc_delta_avg;
}



NTSTATUS bsrDeviceIoControl( PDEVICE_OBJECT DeviceObject, PIRP Irp )
  {
    NTSTATUS sts = STATUS_SUCCESS;
    PIO_STACK_LOCATION pIoStackIrp = NULL;
    PCHAR pInBuf = NULL, pOutBuf = NULL;
    ULONG dwBytesWritten = 0;
    ULONG rsb_timing = 0, threshold = 0;

    DbgPrint("[bsr] > IOCTL handler\n");

    pIoStackIrp = IoGetCurrentIrpStackLocation( Irp );    

    if( pIoStackIrp )
      {
        switch( pIoStackIrp->Parameters.DeviceIoControl.IoControlCode )
          {
          case IOCTL_BUFFIO_RSB_DETECT:

            DbgPrint("[bsr] : got IOCTL_BUFFIO_RSB_DETECT\n");
            if( Irp->AssociatedIrp.SystemBuffer )
              {
                pInBuf = Irp->AssociatedIrp.SystemBuffer;
                pOutBuf = Irp->AssociatedIrp.SystemBuffer;

                if( pIoStackIrp->Parameters.DeviceIoControl.InputBufferLength < sizeof(ULONG) )
                  {
                    DbgPrint( "[bsr] ERROR: STATUS_INVALID_PARAMETER\n" );
                    sts = STATUS_INVALID_PARAMETER;
                    break;
                  }

                threshold = ((ULONG*)pInBuf)[0];
                DbgPrint( "[bsr] : threshold = %u\n", threshold );
                if( !threshold ) threshold = _THRESHOLD_DEFAULT;

                // --
                // -- Luanching "side-channel" detection
                // -- threshold is a minimum value we expect to get when OS is virtualized
                // --
                DbgPrint( "[bsr] : launching RSB side-channel spy..\n");

                rsb_timing = (ULONG)bsr_detect( _MEASURMENTS_COUNT );

                                              DbgPrint( "[bsr] -------------------------------------------------------------------\n" );
                if( rsb_timing >= threshold ) DbgPrint( "[bsr] : VIRTUALIZED (spy measurement >= threshold)\n" );
                else                          DbgPrint( "[bsr] : NOT VIRTUALIZED (spy measurement < threshold)\n" );
                                              DbgPrint( "[bsr] -------------------------------------------------------------------\n" );

                dwBytesWritten = sizeof(ULONG);
                if( pIoStackIrp->Parameters.DeviceIoControl.OutputBufferLength >= sizeof(ULONG) )
                  {
                    ((ULONG*)pInBuf)[ 0 ] = rsb_timing;
                    sts = STATUS_SUCCESS;
                  }
                else
                  {
                    DbgPrint( "[bsr] ERROR: STATUS_BUFFER_TOO_SMALL\n" );
                    sts = STATUS_BUFFER_TOO_SMALL;
                  }
              }
            break;
          }
      }

    Irp->IoStatus.Status = sts;
    Irp->IoStatus.Information = dwBytesWritten;

    DbgPrint("[bsr] < IOCTL handler\n");

    IoCompleteRequest( Irp, IO_NO_INCREMENT );
    return sts;
  }

void DriverUnload( PDRIVER_OBJECT pdobj )
  {
    UNICODE_STRING symlink;
    unsigned int sts = STATUS_SUCCESS;
    DbgPrint( "[bsr] unloading..\n" );

    RtlInitUnicodeString( &symlink, DOS_DEVICE_NAME );
    sts = IoDeleteSymbolicLink( &symlink );
    if( STATUS_SUCCESS != sts )
      DbgPrint( "[bsr] warning: could not delete symbolic link" );

    IoDeleteDevice( pdobj->DeviceObject );
  }

NTSTATUS irp_dispatch( PDEVICE_OBJECT DeviceObject, PIRP Irp )
  {
    Irp->IoStatus.Status = STATUS_SUCCESS;
    IoCompleteRequest( Irp, IO_NO_INCREMENT );
    return STATUS_SUCCESS;
  }


NTSTATUS DriverEntry( PDRIVER_OBJECT pdobj, PUNICODE_STRING RegistryPath ) 
  {
    NTSTATUS sts = STATUS_SUCCESS;
    int i;
    UNICODE_STRING dev_name, dev_dos_name;
    PDEVICE_OBJECT pdvobj;

    DbgPrint( "[bsr] loading..\n" );
    DbgPrint( "[bsr] thread id: %x\n", PsGetCurrentThreadId() );

    for( i = 0; i < IRP_MJ_MAXIMUM_FUNCTION ; i++ )
      {
        pdobj->MajorFunction[ i ] = irp_dispatch;
      }

    pdobj->DriverUnload = DriverUnload;
    pdobj->MajorFunction[ IRP_MJ_DEVICE_CONTROL ] = bsrDeviceIoControl;

    RtlInitUnicodeString( &dev_name, DEVICE_NAME );
    RtlInitUnicodeString( &dev_dos_name, DOS_DEVICE_NAME );

    sts = IoCreateDevice( pdobj, 0, &dev_name, FILE_DEVICE_UNKNOWN, FILE_DEVICE_SECURE_OPEN, FALSE, &pdvobj );
    if( !NT_SUCCESS( sts ) )
      {
        DbgPrint("[bsr] ERROR: failed to create IO device");
        return sts;
      }

    pdvobj->Flags |= DO_BUFFERED_IO;

    sts = IoCreateSymbolicLink( &dev_dos_name, &dev_name );
    if( !NT_SUCCESS( sts ) )
      {
        IoDeleteDevice( pdvobj );
        DbgPrint("[bsr] ERROR: failed to create symbolic link");
        return sts;
      }

    return STATUS_SUCCESS; 
  }
