/*******************************************************************************
 *
 * hyper-channel: implements CPU side-channel based hypervisor detection.
 *
 * Uses RSB side-channel spy that puts Return Stack Buffer (RSB) structure in a known state
 * and monitors its possible corruption by a hyperviser during VMExit.
 *
 * For educational purposes only. Use it on your onw risk.
 *
 * 04/19/2008 c7zero@c7zero.info
 *            www.c7zero.info
 *
 ******************************************************************************/

#include <windows.h>
#include <winioctl.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
//#include <string.h>

#define FILE_DEVICE_BSR 32771

#define IOCTL_BUFFIO_RSB_DETECT \
        CTL_CODE( FILE_DEVICE_BSR, 0x801, METHOD_BUFFERED, FILE_ANY_ACCESS )

#define LEN(s) (sizeof(s)/sizeof(s[0]))

#define DriverName     "bsr"
#define DriverFileName _T("\\bsr.sys")
#define DeviceName     "\\\\.\\bsr"

HANDLE hDevice;
DWORD err = 0;

void rsb_detect( ULONG threshold )
{
  ULONG measurment = 0;
  DWORD lpBytesReturned = 0;

  printf( "[hyper-channel] loading jump program (sending IOCTL_BUFFIO_RSB_DETECT = %#x)..\n", IOCTL_BUFFIO_RSB_DETECT );

  do
    {
      hDevice = CreateFile( DeviceName, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL );
      if( INVALID_HANDLE_VALUE == hDevice )
        {
          printf ( "[hyper-channel] ERROR: CreateFile failed : %d\n", GetLastError() );
          break;
        }

      printf( "[hyper-channel] RSB threshold  : %d\n", threshold );
      DeviceIoControl( hDevice,
                       IOCTL_BUFFIO_RSB_DETECT, 
                       &threshold, 
                       sizeof(ULONG), 
                       &measurment, 
                       sizeof(ULONG), 
                       &lpBytesReturned, 
                       NULL );
      printf( "[hyper-channel] RSB measurement: %d\n", measurment );
      if( measurment >= threshold ) printf( "[hyper-channel] You are IN MATRIX (or a false positive ;)\n" );
      else                          printf( "[hyper-channel] Welcome to the desert of the real. You are NOT IN MATRIX (or a false negative ;)\n" );
    }
  while( 0 );
  if( hDevice ) CloseHandle( hDevice );
}


VOID _cdecl main( ULONG argc, PCHAR argv[] )
{
  SC_HANDLE hscm = NULL, hs = NULL;
  SERVICE_STATUS ss;
  TCHAR driver_path[ MAX_PATH + LEN(DriverFileName) + 1 ] = TEXT("");

  ULONG threshold;

  printf( "\n---------------------------------------------------------------------------------------\n"
          "hyper-channel: implements CPU side-channel based hypervisor detection.\n"
          "(c) 2008 c7zero\n\n"
          "Uses RSB side-channel spy that puts Return Stack Buffer (RSB) structure in a known state\n"
          "and monitors its possible corruption by a hyperviser during VMExit.\n"
          "----------------------------------------------------------------------------------------\n\n" );


  if( argc != 2 )
    {
      printf( "usage: hyper-channel <threshold>\n"
              "<threshold>\tmax # of clock cycles taken by 16 RET'urns on non-virtualized OS\n\n"
              "Note: you are about to load device driver - on your own risk.\n\n" );
      return;
    }
  
  do
    {
      GetCurrentDirectory( MAX_PATH, driver_path );
      _tcscat( driver_path, DriverFileName );

      threshold = strtoul( argv[ 1 ], NULL, 10 );

      hscm = OpenSCManager( NULL, NULL, SC_MANAGER_CREATE_SERVICE ); //SC_MANAGER_ALL_ACCESS

      _tprintf( TEXT("[hyper-channel] loading hyper-channel driver \"%.256s\"..\n"), driver_path );
      if( !hscm )
         {
           printf ( "[hyper-channel] ERROR: OpenSCManager failed with error %d\n", GetLastError() );
           break;
         }
      hs = CreateService( hscm,
                          DriverName,
                          "I AM A SIDE-CHANNEL IN YOUR MARTIX",
                          SERVICE_ALL_ACCESS,
                          SERVICE_KERNEL_DRIVER,
                          SERVICE_DEMAND_START,
                          SERVICE_ERROR_NORMAL,
                          driver_path,
                          NULL, NULL, NULL, NULL, NULL );

      if( !hs ) hs = OpenService( hscm, DriverName, SERVICE_ALL_ACCESS );
      if( !hs )
        {
          printf ( "[hyper-channel] ERROR: OpenService failed with error %d\n", GetLastError() );
          break;
        }

      StartService( hs, 0, NULL );
			
      // --
      // -- call driver IOCTL to start RSB side-channel measurements
      // -- 
      rsb_detect( threshold );

      printf( "[hyper-channel] unloading driver..\n" );

      ControlService( hs, SERVICE_CONTROL_STOP, &ss );
      DeleteService( hs );
      CloseServiceHandle( hs );
    }
  while( 0 );

  if( hscm ) CloseServiceHandle( hscm );
  
}